package com.craigl;
/*
 * Abstract Art Generation Program
 * 
 * This program generates a PNG image of an abstract nature. Something
 * similar to a water color abstract painting. The size and the content
 * of the generated image is controlled by the small number of parameters
 * described below. These parameters interact in strange and unexpected 
 * ways which I will attempt to explain.
 * 
 * WIDTH, HEIGHT - These control the size of the generated image. They
 * are one quarter of the resultant image size in each dimension. If set
 * to 400 and 400 respectively, the resultant image will be 1600x1600.
 * The code was structure this way to speed image generation and to reduce
 * memory requirements.
 * 
 * NUMCOLORS - The color palettes used in these images are generated
 * algorithmically. They are generated by three nested loops. In the case
 * of the RGB palette, the blue component is the outer loop, green in the
 * middle loop and red is the inner loop. Each loop runs from 0 to NUMCOLORS
 * but the actual number of colors in the generated palette is COLOR_COUNT.
 * For an even distribution of colors, NUMCOLORS should be set to the cube
 * root of WIDTH * HEIGHT. However if you want an image that skew towards
 * reds and green while using the RGB palette you can increase NUM_COLORS
 * so that the majority of colors in the palette are reds and greens since
 * palette colors will mostly be generated in the middle and inner loops.
 * If you desire an image with more blues and greens, select the BGR palette
 * and increase the value of NUMCOLORS so more of the blue and green colors
 * ends up in the palette.
 * 
 * COLOR_COUNT - This controls the size of the generated palette and image
 * generation won't be complete until all colors in a selected palette are used. 
 * Algorithm 0 won't generate a full image (parts of the image will be black)
 * if the color count is less than WIDTH * HIEGHT. Algorithm 1 will mostly 
 * generate a complete image but the image will have black trails through it.
 * Very artistic I think.
 * 
 * PALETTE_NUM - This parameter selects which palette to use for image
 * generation. 0 selects the RGB palette, 1 selects the RBG palette,
 * 2 selects the BGR palette, 3 selects the BRG palette, 4 selects the
 * GRB palette and 5 selects the GBR palette. If set to -1, palette selection
 * will be random.
 * 
 * ALGORITHM_NUM - Selects either algorithm 0 or algorithm 1 for image
 * generation. The algorithm selected makes subtle differences in the
 * generated images. If set to -1, the selection of the algorithm will
 * be randomly selected from the two defined.
 * 
 * Playing around with these parameters is the best way to understand
 * how they work and how they interact.
 * 
 * Concept, Design and Implementation: Craig A. Lindley
 * Last Update: 08/16/2020 
 */

import java.awt.Point;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import javax.imageio.ImageIO;

public class AbstractArtWork {

	public class RGB {
		public float red;
		public float green;
		public float blue;
	}

	// Modifiable parameters affecting generated image
	static final int WIDTH = 400;
	static final int HEIGHT = 400;

	static final int NUMCOLORS = 500;
	static final int COLOR_COUNT = 155000;
	static final int PALETTE_NUM = 1;
	static final int ALGORITHM_NUM = 1;

	Randomize rand;

	int algorithm;
	int currentColorIndex;

	// Initialize program storage
	int[] palette = new int[COLOR_COUNT];
	int[][] surface = new int[WIDTH][HEIGHT];
	boolean[][] hasColor = new boolean[WIDTH][HEIGHT];
	boolean[][] isAvailable = new boolean[WIDTH][HEIGHT];

	/********** Palette Generation Methods **********/

	void createRGBPalette() {
		System.out.println("palette: RGB");
		int i = 0;

		for (int b = 0; b < NUMCOLORS; b++) {
			for (int g = 0; g < NUMCOLORS; g++) {
				for (int r = 0; r < NUMCOLORS; r++) {
					int red = r * 255 / (NUMCOLORS - 1);
					int green = g * 255 / (NUMCOLORS - 1);
					int blue = b * 255 / (NUMCOLORS - 1);
					palette[i] = (red << 16) + (green << 8) + blue;

					i++;
					if (i == COLOR_COUNT) {
						return;
					}
				}
			}
		}
	}

	void createRBGPalette() {
		System.out.println("palette: RBG");
		int i = 0;

		for (int g = 0; g < NUMCOLORS; g++) {
			for (int b = 0; b < NUMCOLORS; b++) {
				for (int r = 0; r < NUMCOLORS; r++) {
					int red = r * 255 / (NUMCOLORS - 1);
					int green = g * 255 / (NUMCOLORS - 1);
					int blue = b * 255 / (NUMCOLORS - 1);
					palette[i] = (red << 16) + (green << 8) + blue;

					i++;
					if (i == COLOR_COUNT) {
						return;
					}
				}
			}
		}
	}

	void createBGRPalette() {
		System.out.println("palette: BGR");
		int i = 0;

		for (int r = 0; r < NUMCOLORS; r++) {
			for (int g = 0; g < NUMCOLORS; g++) {
				for (int b = 0; b < NUMCOLORS; b++) {
					int red = r * 255 / (NUMCOLORS - 1);
					int green = g * 255 / (NUMCOLORS - 1);
					int blue = b * 255 / (NUMCOLORS - 1);
					palette[i] = (red << 16) + (green << 8) + blue;

					i++;
					if (i == COLOR_COUNT) {
						return;
					}
				}
			}
		}
	}
	
	void createBRGPalette() {
		System.out.println("palette: BRG");
		int i = 0;

		for (int g = 0; g < NUMCOLORS; g++) {
			for (int r = 0; r < NUMCOLORS; r++) {
				for (int b = 0; b < NUMCOLORS; b++) {
					int red = r * 255 / (NUMCOLORS - 1);
					int green = g * 255 / (NUMCOLORS - 1);
					int blue = b * 255 / (NUMCOLORS - 1);
					palette[i] = (red << 16) + (green << 8) + blue;

					i++;
					if (i == COLOR_COUNT) {
						return;
					}
				}
			}
		}
	}
	
	void createGRBPalette() {
		System.out.println("palette: GRB");
		int i = 0;

		for (int b = 0; b < NUMCOLORS; b++) {
			for (int r = 0; r < NUMCOLORS; r++) {
				for (int g = 0; g < NUMCOLORS; g++) {
					int red = r * 255 / (NUMCOLORS - 1);
					int green = g * 255 / (NUMCOLORS - 1);
					int blue = b * 255 / (NUMCOLORS - 1);
					palette[i] = (red << 16) + (green << 8) + blue;

					i++;
					if (i == COLOR_COUNT) {
						return;
					}
				}
			}
		}
	}
	
	void createGBRPalette() {
		System.out.println("palette: GBR");
		int i = 0;

		for (int r = 0; r < NUMCOLORS; r++) {
			for (int b = 0; b < NUMCOLORS; b++) {
				for (int g = 0; g < NUMCOLORS; g++) {
					int red = r * 255 / (NUMCOLORS - 1);
					int green = g * 255 / (NUMCOLORS - 1);
					int blue = b * 255 / (NUMCOLORS - 1);
					palette[i] = (red << 16) + (green << 8) + blue;

					i++;
					if (i == COLOR_COUNT) {
						return;
					}
				}
			}
		}
	}
	
	void shufflePalette() {
		for (int a = 0; a < COLOR_COUNT; a++) {
			int r = rand.randomIntNtoM(a, COLOR_COUNT);
			int temp = palette[a];
			palette[a] = palette[r];
			palette[r] = temp;
		}
	}

	void createPalette() {

		int colorSort = rand.randomInt0toN(6);

		if (PALETTE_NUM != -1) {
			colorSort = PALETTE_NUM;
		}

		switch (colorSort) {
		case 0:
			createRGBPalette();
			break;
		case 1:
			createRBGPalette();
			break;
		case 2:
			createBGRPalette();
			break;
		case 3:
			createBRGPalette();
			break;
		case 4:
			createGRBPalette();
			break;
		case 5:
			createGBRPalette();
			break;
		}
		shufflePalette();
	}

	/********** Misc Methods **********/

	int colorDifference(int c1, int c2) {

		int c1Red = (c1 >> 16) & 0xFF;
		int c1Green = (c1 >> 8) & 0xFF;
		int c1Blue = (c1 & 0xFF);

		int c2Red = (c2 >> 16) & 0xFF;
		int c2Green = (c2 >> 8) & 0xFF;
		int c2Blue = (c2 & 0xFF);

		int r = c1Red - c2Red;
		int g = c1Green - c2Green;
		int b = c1Blue - c2Blue;
		return r * r + g * g + b * b;
	}

	void markAvailableNeighbors(Point point) {
		for (int dy = -1; dy <= 1; dy++) {
			int ny = point.y + dy;

			if (ny == -1 || ny == HEIGHT)
				continue;

			for (int dx = -1; dx <= 1; dx++) {
				if (dx == 0 && dy == 0)
					continue;

				int nx = point.x + dx;

				if (nx == -1 || nx == WIDTH)
					continue;

				if (!hasColor[nx][ny]) {
					isAvailable[nx][ny] = true;
				}
			}
		}
	}

	Point getAvailablePointWithClosestNeighborColor(int color) {

		Point best = new Point();

		// find the pixel with the smallest difference between the current color and all
		// of it's neighbors' colors
		int smallestDifference = 999999;
		for (int y = 0; y < HEIGHT; y++) {
			for (int x = 0; x < WIDTH; x++) {
				// skip any that arent' available
				if (!isAvailable[x][y]) {
					continue;
				}

				// loop through its neighbors
				int smallestDifferenceAmongNeighbors = 999999;
				for (int dy = -1; dy <= 1; dy++) {
					if (y + dy == -1 || y + dy == HEIGHT) {
						continue;
					}

					for (int dx = -1; dx <= 1; dx++) {
						if (x == 0 && y == 0) {
							continue;
						}

						if (x + dx == -1 || x + dx == WIDTH) {
							continue;
						}

						int nx = x + dx;
						int ny = y + dy;

						// skip any neighbors that don't already have a color
						if (!hasColor[nx][ny]) {
							continue;
						}

						int neighborColor = surface[nx][ny];

						int difference = colorDifference(neighborColor, color);

						if (difference < smallestDifferenceAmongNeighbors
								|| (difference == smallestDifferenceAmongNeighbors && rand.randomInt0toN(2) == 1)) {
							smallestDifferenceAmongNeighbors = difference;
						}
					}
				}

				if (smallestDifferenceAmongNeighbors < smallestDifference
						|| (smallestDifferenceAmongNeighbors == smallestDifference && rand.randomInt0toN(2) == 1)) {
					smallestDifference = smallestDifferenceAmongNeighbors;
					best.x = x;
					best.y = y;
				}
			}
		}
		return best;
	}

	Point getAvailablePointWithClosestAverageNeighborColor(int color) {

		Point best = new Point();

		int smallestAverageDifference = 999999;

		for (int y = 0; y < HEIGHT; y++) {
			for (int x = 0; x < WIDTH; x++) {
				// skip any that arent' available
				if (!isAvailable[x][y]) {
					continue;
				}
				int neighborCount = 0;
				int neighborColorDifferenceTotal = 0;

				// loop through its neighbors
				for (int dy = -1; dy <= 1; dy++) {
					if (y + dy == -1 || y + dy == HEIGHT) {
						continue;
					}

					for (int dx = -1; dx <= 1; dx++) {
						if (x + dx == -1 || x + dx == WIDTH) {
							continue;
						}

						int nx = x + dx;
						int ny = y + dy;

						// skip any neighbors that don't already have a color
						if (!hasColor[nx][ny]) {
							continue;
						}
						neighborCount++;

						int neighborColor = surface[nx][ny];

						int difference = colorDifference(neighborColor, color);

						neighborColorDifferenceTotal += difference;
					}
				}

				int averageDifferenceAmongNeighbors = neighborColorDifferenceTotal / neighborCount;

				if (averageDifferenceAmongNeighbors < smallestAverageDifference
						|| (averageDifferenceAmongNeighbors == smallestAverageDifference
								&& rand.randomInt0toN(2) == 1)) {
					smallestAverageDifference = averageDifferenceAmongNeighbors;
					best.x = x;
					best.y = y;
				}
			}
		}
		return best;
	}

	Point getAvailablePoint(int algorithm, int color) {
		switch (algorithm) {
		case 0:
			return getAvailablePointWithClosestNeighborColor(color);
		case 1:
		default:
			return getAvailablePointWithClosestAverageNeighborColor(color);
		}
	}

	/********** Water Color Generation Method **********/

	private void generateWaterColor(BufferedImage bi) {

		currentColorIndex = 0;

		// Create a palette
		createPalette();

		algorithm = rand.randomInt0toN(2);
		if (ALGORITHM_NUM != -1) {
			algorithm = ALGORITHM_NUM;
		}
		System.out.printf("Algorithm: %d\n", algorithm);

		// clear all flags
		for (int y = 0; y < HEIGHT; y++) {
			for (int x = 0; x < WIDTH; x++) {
				hasColor[x][y] = false;
				isAvailable[x][y] = false;
				surface[x][y] = 0;
			}
		}

		Point point = new Point();

		do {
			int color = palette[currentColorIndex];

			if (currentColorIndex == 0) {
				// use a random starting point
				point.x = rand.randomInt0toN(WIDTH);
				point.y = rand.randomInt0toN(HEIGHT);
			} else {
				point = getAvailablePoint(algorithm, color);
			}

			isAvailable[point.x][point.y] = false;
			hasColor[point.x][point.y] = true;

			// Log the surface color
			surface[point.x][point.y] = color;

			// Expand pixel into block to cover whole display
			int x = 4 * point.x;
			int y = 4 * point.y;

			bi.setRGB(x, y, color);
			bi.setRGB(x + 1, y, color);
			bi.setRGB(x + 2, y, color);
			bi.setRGB(x + 3, y, color);
			bi.setRGB(x, y + 1, color);
			bi.setRGB(x + 1, y + 1, color);
			bi.setRGB(x + 2, y + 1, color);
			bi.setRGB(x + 3, y + 1, color);
			bi.setRGB(x, y + 2, color);
			bi.setRGB(x + 1, y + 2, color);
			bi.setRGB(x + 2, y + 2, color);
			bi.setRGB(x + 3, y + 2, color);
			bi.setRGB(x, y + 3, color);
			bi.setRGB(x + 1, y + 3, color);
			bi.setRGB(x + 2, y + 3, color);
			bi.setRGB(x + 3, y + 3, color);

			markAvailableNeighbors(point);

			currentColorIndex++;

		} while (currentColorIndex < COLOR_COUNT);
	}

	private void run(int width, int height) throws IOException {

		// create instance of Randomize class
		rand = new Randomize();

		// Constructs a BufferedImage of type RGB
		// The buffered image is 4x the specified dimension
		// because pixels are 4x4 blocks
		BufferedImage bufferedImage = new BufferedImage(width * 4, height * 4, BufferedImage.TYPE_INT_RGB);

		// Generate the watercolor data
		generateWaterColor(bufferedImage);

		// Save as PNG
		File file = new File("/Users/craiglindley/Documents/images/AbstractArt/abstract_art.png");
		ImageIO.write(bufferedImage, "png", file);
	}

	/********** Program Entry Point **********/

	public static void main(String[] args) throws IOException {

		AbstractArtWork aaw = new AbstractArtWork();
		aaw.run(WIDTH, HEIGHT);
	}
}
